/************************************************************************
     File:        MyArcBall.cpp

     Author:     
                  Michael Gleicher, gleicher@cs.wisc.edu
     Modifier
                  Yu-Chi Lai, yu-chi@cs.wisc.edu
     
     Comment:    
									Written October, 2001, Michael L. Gleicher
									Updated October, 2005 for CS559 - Michael L. Gleicher

		 
									An arcball is a user interface for rotating an object around
									its very useful for looking around the world.

									The actual mathematics to implement it are elegant, but beyond the
									scope of CS559. This example code is provided for you to use
									without having to worry about the details of how it works.

		To use this code:
								1. Create an instance of the "ArcBall" in your window class
								2. When you draw, pull that matrix out of the ArcBall (using the
								    getMatrix method).
								3. When you get a mouse button push event, figure out where the event
								   happened (in normalized device coordinates) and call the "down"
								   method of the ArcBall
								4. When you get a drag event (and the button that is pushed down is
								   the one for moving the viewpoint)
    Note:
								1.	A quickly hacked together implementation of ArcBall, with lots of
									  bits lifted from Ken Shoemake's original source in Graphics Gems
									  made into a nice C++ thing

								2.	It is simplified a lot - the controller is always at the origin,
								3.	It has a fixed radius, it doesn't have constraints


   
     Platform:    Visio Studio.Net 2003
*************************************************************************/

#include "MyArcBall.H"
#include <math.h>

#include <windows.h>
#include <GL/gl.h>

#include "stdio.h"

enum QuatPart {X=0, Y=1, Z=2, W=3, QuatLen};

//**************************************************************************
//
// minimal quaternion class
//==========================================================================
Quat::Quat(float ix, float iy, float iz, float iw) 
	:  x(ix), y(iy), z(iz), w(iw)
//==========================================================================
{
}
//**************************************************************************
//
//==========================================================================
Quat::Quat() 
	:  x(0), y(0), z(0), w(1)
//==========================================================================
{
}
//**************************************************************************
//
//==========================================================================
Quat::Quat(const Quat& q) 
	:  x(q.x), y(q.y), z(q.z), w(q.w)
//==========================================================================
{
}

//**************************************************************************
//
// Renormalize, in case things got messed up
//==========================================================================
void Quat::renorm()
//==========================================================================
{
    float Nq = 1.f / (float) sqrt(x*x + y*y + z*z + w*w);
	x *= Nq;
	y *= Nq;
	z *= Nq;
	w *= Nq;
}

//**************************************************************************
//
// Conversions
//==========================================================================
void Quat::toMatrix(HMatrix out) const
//==========================================================================
{
    float Nq = x*x + y*y + z*z + w*w;
    float s = (Nq > 0.f) ? (2.0f / Nq) : 0.f;
    float xs = x*s,	      ys = y*s,	  zs = z*s;
    float wx = w*xs,	  wy = w*ys,  wz = w*zs;
    float xx = x*xs,	  xy = x*ys,  xz = x*zs;
    float yy = y*ys,	  yz = y*zs,  zz = z*zs;
    out[X][X] = 1.0f - (yy + zz); out[Y][X] = xy + wz;          out[Z][X] = xz - wy;
    out[X][Y] = xy - wz;          out[Y][Y] = 1.0f - (xx + zz); out[Z][Y] = yz + wx;
    out[X][Z] = xz + wy;          out[Y][Z] = yz - wx;          out[Z][Z] = 1.0f - (xx + yy);
    out[X][W] = out[Y][W] = out[Z][W] = out[W][X] = out[W][Y] = out[W][Z] = 0.0f;
    out[W][W] = 1.0f;
}

//**************************************************************************
//
// quaternion operations
//==========================================================================
Quat Quat::conjugate() const
//==========================================================================
{
  return Quat(-x,-y,-z,w);
}

//**************************************************************************
//
//==========================================================================
Quat Quat::operator* (const Quat& qR) const
//==========================================================================
{
	Quat qq;
	qq.w = w*qR.w - x*qR.x - y*qR.y - z*qR.z;
	qq.x = w*qR.x + x*qR.w + y*qR.z - z*qR.y;
	qq.y = w*qR.y + y*qR.w + z*qR.x - x*qR.z;
	qq.z = w*qR.z + z*qR.w + x*qR.y - y*qR.x;
	return (qq);
}


//**************************************************************************
//
//==========================================================================
MyArcBall::MyArcBall()
//==========================================================================
{
}

//**************************************************************************
//
// when the mouse goes down, remember where. also, clear out the now
// position by putting it into start
//==========================================================================
void MyArcBall::down(const float x, const float y)
//==========================================================================
{
	start = now * start;
	now = Quat();		// identity

	downX = x;
	downY = y;
}

//**************************************************************************
//
// get the whole matrix!
//==========================================================================
void MyArcBall::getMatrix(HMatrix m) const
//==========================================================================
{
	Quat qAll = now * start;
	qAll = qAll.conjugate();   // since Ken does everything transposed
	qAll.toMatrix(m);
}

//**************************************************************************
//
// a simplified interface - so you never see the insides of arcball
//==========================================================================
void MyArcBall::multMatrix()
//==========================================================================
{
	HMatrix m;
	getMatrix(m);
    glMultMatrixf((float*) m);
}
//**************************************************************************
//
//==========================================================================
void MyArcBall::reset(float x, float y, float z)
//==========================================================================
{
	start.x = x;
	start.y = y;
	start.z = z;
	float iw = x*x + y*y + z*z;
	if (iw<1)
		start.w = sqrt(1-iw);
	else
		start.w = 0;
	start.renorm();

	now.x = now.y = now.z = 0; now.w = 1;
}

//**************************************************************************
//
//==========================================================================
void MyArcBall::spin(float x, float y, float z)
//==========================================================================
{
	printf("\nstart was %g %g %g %g\n",start.x,start.y,start.z,start.w);
	// first, get rid of anything cached
	start = now * start;
	now = Quat();

	float iw = x*x + y*y + z*z;
	if (iw<1)
		iw = sqrt(1-iw);
	else
		iw = 0;
	Quat newq(x,y,z,iw);

	newq.renorm();
	start = newq * start;
	printf(" (no norm) %g %g %g %g\n",start.x,start.y,start.z,start.w);
	start.renorm();
	printf("start is %g %g %g %g\n",start.x,start.y,start.z,start.w);
}

//**************************************************************************
//
// the gory guts

//**************************************************************************
//
// a helper - figure out where in XYZ a mouse down goes
// assumes sphere of unit radius
//==========================================================================
static void onUnitSphere(const float mx, const float my,
													float& x, float& y, float& z)
//==========================================================================
{
	x = mx;		// should divide radius
	y = my;
	float mag = x*x + y*y;
	if (mag > 1.0f) {
		float scale = 1.0f / ((float) sqrt(mag));
		x *= scale;
		y *= scale;
		z = 0;
	} else {
		z = (float) sqrt(1 - mag);
	}
}

//**************************************************************************
//
//==========================================================================
void MyArcBall::computeNow(const float nowX, const float nowY)
//==========================================================================
{
	float dx,dy,dz;
	float mx,my,mz;
	onUnitSphere(downX, downY, dx, dy, dz);
	onUnitSphere(nowX, nowY, mx, my, mz);

	// here we compute the quaternion between these two points
  now.x = dy*mz - dz*my;
  now.y = dz*mx - dx*mz;
  now.z = dx*my - dy*mx;
  now.w = dx*mx + dy*my + dz*mz;

	now.renorm();		// just in case...
}
