#include <Fl/Fl.h>
#include <Fl/Gl.h>
#include <Gl/Glu.h>
#include <FL/Fl_Gl_Window.h>
#include "MyWindow.h"
#include "3DUtils.h"
#include "Cube.h"

/************************************************************************
     File:       MyWindow.cpp

     Author:     Steven J Chenney, schenney@cs.wisc.edu
     Modifier
                  Yu-Chi Lai, yu-chi@cs.wisc.edu
				  Justin Fiedler
     
     Comment:     Simple window class with nothing in it
				  Justin Fiedler Additions:
					A window class that handles OpenGL.  Holds a "world" with a checkerboard and
					cubes.  Various keyboard events allow user interaction:
						A: Adds a new Cube at the orgin.
						C: Toggles between moving the selected cube and moving the camera.
						S: Toggle the selected cube.
						T: Toggles the plane of movement. Default is the XZ plane, alternative is
							the XY plane.
						Arrow keys: Left/Right move in the X direction. Up/Down move in the Z/Y direction
							depending on the selected plane of movement.
						Spacebar: Animated the selected cube by sending it into a vicious spin!
   
     Platform:    Visio Studio.Net 2003
*************************************************************************/

//Idle callback for animation that rotates the select cube
void IdleCallback(void* pData)
{	if (pData != NULL)
	{	MyWindow* pWindow = reinterpret_cast<MyWindow*>(pData);
		if (pWindow->animating) {
			pWindow->rotation += pWindow->rotationIncrement;
			pWindow->redraw();
		}
	}
}

//**************************************************************************************
//
// * Constructor
//======================================================================================
MyWindow::MyWindow(int x, int y, int w, int h) :  Fl_Gl_Window(x,y,w,h)
//======================================================================================
{	mode(FL_RGB | FL_ALPHA | FL_DEPTH | FL_DOUBLE);
	moveInc = .1;
	selected = 0;
	cubes.push_back(Cube(0, .5, 0));
	camera = new Cube(4, 4, 4);
	rotation = 0.f;
	rotationIncrement =5.f;
	animating = false;
	planeXZ = true;
	cameraMove = false;
	Fl::add_idle(IdleCallback, this);

}

//**************************************************************************************
//
// * Destructor
//======================================================================================
MyWindow::~MyWindow(void)
//======================================================================================
{}

//first time setup of the world
void MyWindow::InitializeGL()
{
	glClearColor(.1f, .1f, .1f, 1);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	float lightColor[4]= {1, 1, 1, 1};
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightColor);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

}

//the overloaded draw function
void MyWindow::draw()
{
	static bool firstTime = true;
	if (firstTime) {
		InitializeGL();
		firstTime = false;
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // clear the color and depth buffer
	// view transformations
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1, 1, -1, 1, 1, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	//move camera
	gluLookAt(camera->x, camera->y, camera->z, 0, 0, 0, 0, 1, 0);
	
	//shine a light
	float lightPosition[4] = {5, 5, 5, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
	
	//ground is good
	setupFloor();
	drawFloor();

	//and some shadows
	setupShadows();
	drawCubes(0);
	unsetupShadows();
	
	//and finally the cubes
	drawCubes(1);
}

//handle keyboard events
int MyWindow::handle(int event)
{	Cube* c;
	if(cameraMove){
		c = camera;
	}else{
		c = &cubes[selected];
	}
	switch (event)	{
		case FL_FOCUS:
		case FL_UNFOCUS:
			return 1;

		case FL_KEYBOARD:
			int key = Fl::event_key();
			switch (key) {
				case 'a':		//ADD A NEW CUBE AT THE ORGIN
				case 'A':
					cubes.push_back(Cube(0, .5, 0));
					redraw();
					break;
				case 'c':		//TOGGLE CAMERA/SELECTED CUBE
				case 'C':
					cameraMove = !cameraMove;
					break;
				case 's':		//TOGGLE SELECTED CUBE
				case 'S':
					selected++;
					if(selected >= cubes.size()) selected = 0;
					redraw();
					break;
				case 't':		//TOGGLE PLANE OF MOVEMENT
				case 'T':
					planeXZ = !planeXZ;
					break;
				case FL_Left:	//MOVEMENT KEYS
					moveCube(c, -moveInc, 0, 0);
					redraw();
					return 1;
  				case FL_Right:
					moveCube(c, moveInc, 0, 0);
					redraw();
					return 1;
				case FL_Up:
					if(planeXZ) {
						moveCube(c, 0, 0, -moveInc);
					} else {
						moveCube(c, 0, moveInc, 0);
					}
					redraw();
					return 1;
  				case FL_Down:
					if(planeXZ){
						moveCube(c, 0, 0, moveInc);
					} else {
						moveCube(c, 0, -moveInc, 0);
					}
					redraw();
					return 1;
				case ' ':
					animating = !animating;
					return 1;
				}
		 }

	 return Fl_Gl_Window::handle(event);
}

//draw the cubes
void MyWindow::drawCubes(int doColors)
{	for(int i = 0; i < (int) cubes.size(); i++) {
		glPushMatrix();
		if (doColors) {
			if (selected == i){
				glColor3f(1,1,0);
			}else{
				glColor3f(1,0,0);
			}
		}
		//move cube to current location
		glTranslatef(cubes[i].x, cubes[i].y, cubes[i].z);
		//if animated rotate selected cube
		if(animating && selected == i){
			glRotatef(rotation, 0, 1, 0);
		}
		//finally draw the cube
		drawCube(0, 0, 0);
		glPopMatrix();
	}
}